@@ -67,6 +67,8 @@ our %MIME_TYPES = (
     json => 'application/json',
     xml  => 'application/xml',
     txt  => 'text/plain',
+    pl   => 'text/plain',
+    md   => 'text/plain',
     png  => 'image/png',
     jpg  => 'image/jpeg',
     jpeg => 'image/jpeg',
